@@ -30,7 +30,9 @@ enum {
 	EOG_IMAGE_LOAD_STATUS_DONE     = 1 << 2,
 	EOG_IMAGE_LOAD_STATUS_FAILED   = 1 << 3,
 	EOG_IMAGE_LOAD_STATUS_CANCELLED = 1 << 4,
-	EOG_IMAGE_LOAD_STATUS_INFO_DONE = 1 << 5
+	EOG_IMAGE_LOAD_STATUS_INFO_DONE = 1 << 5,
+	EOG_IMAGE_LOAD_STATUS_TRANSFORMED = 1 << 6,
+	EOG_IMAGE_LOAD_STATUS_PROGRESS = 1 << 7
 };
 
 struct _EogImagePrivate {
@@ -58,6 +60,8 @@ struct _EogImagePrivate {
 	GMutex *status_mutex;
 	gint load_status;
 	gboolean cancel_loading;
+	float progress; /* Range from [0.0...1.0] indicate the progress of 
+			   actions in percent */
 
 	/* data which depends on the load status */
 	int update_x1;
@@ -65,6 +69,11 @@ struct _EogImagePrivate {
 	int update_x2;
 	int update_y2;
 	char *error_message;
+	
+	/* stack of transformations recently applied */
+	GList *undo_stack;
+	/* composition of all applied transformations */
+	EogTransform *trans;
 };
 
 enum {
@@ -74,6 +83,7 @@ enum {
 	SIGNAL_LOADING_INFO_FINISHED,
 	SIGNAL_LOADING_FAILED,
 	SIGNAL_LOADING_CANCELLED,
+	SIGNAL_PROGRESS,
 	SIGNAL_IMAGE_CHANGED,
 	SIGNAL_THUMBNAIL_FINISHED,
 	SIGNAL_THUMBNAIL_FAILED,
@@ -302,6 +312,7 @@ static void
 eog_image_dispose (GObject *object)
 {
 	EogImagePrivate *priv;
+	GList *it;
 
 	priv = EOG_IMAGE (object)->priv;
 
@@ -330,6 +341,20 @@ eog_image_dispose (GObject *object)
 		priv->status_mutex = NULL;
 	}
 
+	if (priv->trans) {
+		g_object_unref (priv->trans);
+		priv->trans = NULL;
+	}
+
+	if (priv->undo_stack) {
+		for (it = priv->undo_stack; it != NULL; it = it->next){
+			g_object_unref (G_OBJECT (it->data));
+		}
+
+		g_list_free (priv->undo_stack);
+		priv->undo_stack = NULL;
+	}
+
 #if HAVE_EXIF
 	if (priv->exif) {
 		exif_data_unref (priv->exif);
@@ -421,6 +446,15 @@ eog_image_class_init (EogImageClass *klass)
 			      NULL, NULL,
 			      g_cclosure_marshal_VOID__VOID,
 			      G_TYPE_NONE, 0);
+	eog_image_signals [SIGNAL_PROGRESS] = 
+		g_signal_new ("progress",
+			      G_TYPE_OBJECT,
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (EogImageClass, progress),
+			      NULL, NULL,
+			      g_cclosure_marshal_VOID__FLOAT,
+			      G_TYPE_NONE, 1,
+			      G_TYPE_FLOAT);
 	eog_image_signals [SIGNAL_IMAGE_CHANGED] = 
 		g_signal_new ("image_changed",
 			      G_TYPE_OBJECT,
@@ -489,6 +523,9 @@ eog_image_new_uri (GnomeVFSURI *uri)
 	priv->modified = FALSE;
 	priv->load_thread = NULL;
 
+	priv->undo_stack = NULL;
+	priv->trans = NULL;
+
 #if OBJECT_WATCH
 	n_active_images++;
 	g_message ("active image objects: %i", n_active_images);
@@ -528,6 +565,7 @@ check_load_status (gpointer data)
 	int load_status;
 	int x, y, width, height;
 	gboolean call_again = TRUE;
+	float progress;
 
 	img = EOG_IMAGE (data);
 	priv = img->priv;
@@ -542,6 +580,7 @@ check_load_status (gpointer data)
 	y = priv->update_y1;
 	width = priv->update_x2 - x;
 	height = priv->update_y2 - y;
+	progress = priv->progress;
 
 	priv->load_status = 0;
 	priv->update_x1 = 10000;
@@ -550,7 +589,7 @@ check_load_status (gpointer data)
 	priv->update_y2 = 0;
 
 	g_mutex_unlock (priv->status_mutex);
-	
+
 	if ((load_status & EOG_IMAGE_LOAD_STATUS_FAILED) > 0) {
 		g_signal_emit (G_OBJECT (img), eog_image_signals [SIGNAL_LOADING_FAILED], 0, priv->error_message);
 		call_again = FALSE;
@@ -575,10 +614,19 @@ check_load_status (gpointer data)
 	}
 	
  	if ((load_status & EOG_IMAGE_LOAD_STATUS_DONE) > 0) {
-		g_signal_emit (G_OBJECT (img), eog_image_signals [SIGNAL_LOADING_FINISHED], 0);
+		if ((load_status & EOG_IMAGE_LOAD_STATUS_TRANSFORMED) > 0) {
+			g_signal_emit (G_OBJECT (img), eog_image_signals [SIGNAL_IMAGE_CHANGED], 0);
+		}
+		else {
+			g_signal_emit (G_OBJECT (img), eog_image_signals [SIGNAL_LOADING_FINISHED], 0);
+		}
 		call_again = FALSE;
 	}
 
+ 	if ((load_status & EOG_IMAGE_LOAD_STATUS_PROGRESS) > 0) {
+		g_signal_emit (G_OBJECT (img), eog_image_signals [SIGNAL_PROGRESS], 0, progress);
+	}
+
 	if (call_again) {
 		priv->load_id = g_timeout_add (CHECK_LOAD_TIMEOUT, check_load_status, img);
 	}
@@ -590,6 +638,23 @@ check_load_status (gpointer data)
 }
 
 
+static void
+transform_progress_hook_async (EogTransform *trans, float progress, gpointer hook_data)
+{
+	EogImage *img;
+	EogImagePrivate *priv;
+
+	img = EOG_IMAGE(hook_data);
+	priv = img->priv;
+
+	if (progress != priv->progress) {
+		g_mutex_lock (priv->status_mutex);
+		priv->progress = progress;
+		priv->load_status |= EOG_IMAGE_LOAD_STATUS_PROGRESS;
+		g_mutex_unlock (priv->status_mutex);		
+	}
+}
+
 static void 
 load_area_updated (GdkPixbufLoader *loader, gint x, gint y, gint width, gint height, gpointer data)
 {
@@ -772,10 +837,12 @@ real_image_load (gpointer data)
 	EogImage *img;
 	EogImagePrivate *priv;
 	GdkPixbufLoader *loader;
+	GnomeVFSFileInfo *info;
 	GnomeVFSResult result;
 	GnomeVFSHandle *handle;
 	guchar *buffer;
 	GnomeVFSFileSize bytes_read;
+	GnomeVFSFileSize bytes_read_total;
 	gboolean failed;
 	GError *error = NULL;
 #if HAVE_EXIF
@@ -806,6 +873,25 @@ real_image_load (gpointer data)
 
 		return NULL;
 	}
+
+	/* determine file size */
+	/* FIXME: we should reuse the values gained in eog_image_load here */
+	info = gnome_vfs_file_info_new ();
+	result = gnome_vfs_get_file_info_uri (priv->uri,
+					      info,
+					      GNOME_VFS_FILE_INFO_DEFAULT);
+	if ((result != GNOME_VFS_OK) || (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_SIZE) == 0) {
+		g_mutex_lock (priv->status_mutex);
+		priv->load_status |= EOG_IMAGE_LOAD_STATUS_FAILED;
+		priv->error_message = (char*) gnome_vfs_result_to_string (result);
+		g_mutex_unlock (priv->status_mutex);
+
+		gnome_vfs_file_info_unref (info);
+
+		return NULL;
+	}
+	bytes_read_total = 0;
+	priv->progress = 0.0;
 	
 	buffer = g_new0 (guchar, READ_BUFFER_SIZE);
 	loader = gdk_pixbuf_loader_new ();
@@ -819,8 +905,10 @@ real_image_load (gpointer data)
 		g_signal_connect_object (G_OBJECT (loader), "area-updated", (GCallback) load_area_updated, img, 0);
 		g_signal_connect_object (G_OBJECT (loader), "size-prepared", (GCallback) load_size_prepared, img, 0);
 	}
-	
+
 	while (!priv->cancel_loading) {
+		float new_progress;
+		
 		result = gnome_vfs_read (handle, buffer, READ_BUFFER_SIZE, &bytes_read);
 		if (result == GNOME_VFS_ERROR_EOF || bytes_read == 0) {
 			break;
@@ -834,6 +922,14 @@ real_image_load (gpointer data)
 			failed = TRUE;
 			break;
 		}
+
+		bytes_read_total += bytes_read;
+		new_progress = (float) bytes_read_total / (float) info->size;
+		if (priv->progress != new_progress) {
+			priv->progress = new_progress;
+			priv->load_status |= EOG_IMAGE_LOAD_STATUS_PROGRESS;
+		}
+
 #if HAVE_EXIF
 		if (exif_loader_write (eld, buffer, bytes_read)) {
 			g_mutex_lock (img->priv->status_mutex);
@@ -849,15 +945,16 @@ real_image_load (gpointer data)
 
 	g_free (buffer);
 	gnome_vfs_close (handle);
+	gnome_vfs_file_info_unref (info);
 	
-	g_mutex_lock (priv->status_mutex);
-
 	if (failed) {
+		g_mutex_lock (priv->status_mutex);
 		if (priv->image != NULL) {
 			g_object_unref (priv->image);
 			priv->image = NULL;
 		}
 		priv->load_status |= EOG_IMAGE_LOAD_STATUS_FAILED;
+		priv->progress = 0.0;
 		
 		if (error != NULL) {
 			priv->error_message = g_strdup (error->message);
@@ -865,27 +962,53 @@ real_image_load (gpointer data)
 		else {
 			priv->error_message = NULL;
 		}
+		g_mutex_unlock (priv->status_mutex);
 	}
 	else if (priv->cancel_loading) {
+		g_mutex_lock (priv->status_mutex);
 		if (priv->image != NULL) {
 			g_object_unref (priv->image);
 			priv->image = NULL;
 		}
 		priv->cancel_loading = TRUE;
+		priv->progress = 0.0;
 
 		priv->load_status |= EOG_IMAGE_LOAD_STATUS_CANCELLED;
+		g_mutex_unlock (priv->status_mutex);
 	}
 	else {
-		if (priv->image == NULL) {
-			priv->image = gdk_pixbuf_loader_get_pixbuf (loader);
-			g_assert (priv->image != NULL);
-			g_object_ref (priv->image);
+		GdkPixbuf *image;
+		GdkPixbuf *transformed = NULL;
+		EogTransform *trans;
 
-			priv->width = gdk_pixbuf_get_width (priv->image);
-			priv->height = gdk_pixbuf_get_height (priv->image);
-			priv->load_status |= EOG_IMAGE_LOAD_STATUS_PREPARED;
+		g_mutex_lock (priv->status_mutex);
+		image = priv->image;
+		trans = priv->trans;
+		g_mutex_unlock (priv->status_mutex);
+		
+		if (image == NULL) {
+			image = gdk_pixbuf_loader_get_pixbuf (loader);
+			g_object_ref (image);
+		}
+		g_assert (image != NULL);
+		
+		if (trans != NULL) {
+			transformed = eog_transform_apply (trans, image, transform_progress_hook_async, img);
+
+			g_object_unref (image);
+			image = transformed;
+		}
+
+		g_mutex_lock (priv->status_mutex);
+		priv->progress = 1.0;
+		priv->image = image;
+		priv->width = gdk_pixbuf_get_width (priv->image);
+		priv->height = gdk_pixbuf_get_height (priv->image);
+		if (trans != NULL) {
+			priv->load_status |= EOG_IMAGE_LOAD_STATUS_TRANSFORMED;
 		}
 		priv->load_status |= EOG_IMAGE_LOAD_STATUS_DONE;
+		g_mutex_unlock (priv->status_mutex);
 	}
 
 #if HAVE_EXIF
@@ -895,15 +1018,16 @@ real_image_load (gpointer data)
 		exif_data_unref (eld->data);
 	g_free (eld);
 #endif
-
-	priv->load_thread = NULL;
-	g_mutex_unlock (priv->status_mutex);
 	
 	if (error != NULL) {
 		g_error_free (error);
 	}
 	g_object_unref (loader);
 
+	g_mutex_lock (priv->status_mutex);
+	priv->load_thread = NULL;
+	g_mutex_unlock (priv->status_mutex);
+
 	return NULL;
 }
 
@@ -1054,91 +1178,125 @@ eog_image_get_size (EogImage *img, int *width, int *height)
 	*height = priv->height;
 }
 
-
-void    
-eog_image_rotate_clock_wise (EogImage *img)
+static gboolean
+check_progress_sync (gpointer data)
 {
-	EogImagePrivate *priv;
-	GdkPixbuf *rotated;
-
-	g_return_if_fail (EOG_IS_IMAGE (img));
-
-	priv = img->priv;
-	if (priv->image == NULL) return;
-
-	rotated = eog_pixbuf_rotate_90_cw (priv->image);
-	g_object_unref (priv->image);
-	priv->image = rotated;
-
-	priv->modified = TRUE;
-	g_signal_emit (G_OBJECT (img), eog_image_signals [SIGNAL_IMAGE_CHANGED], 0);
+	g_signal_emit (G_OBJECT (data), eog_image_signals [SIGNAL_PROGRESS], 0, EOG_IMAGE (data)->priv->progress);
+	
+	return TRUE;
 }
 
-void    
-eog_image_rotate_counter_clock_wise (EogImage *img)
+static void
+transform_progress_hook_sync (EogTransform *trans, float progress, gpointer hook_data)
 {
 	EogImagePrivate *priv;
-	GdkPixbuf *rotated;
 
-	g_return_if_fail (EOG_IS_IMAGE (img));
+	priv = EOG_IMAGE (hook_data)->priv;
 
-	priv = img->priv;
-	if (priv->image == NULL) return;
+	if (progress != priv->progress) {
+		
+		priv->progress = progress;
 
-	rotated = eog_pixbuf_rotate_90_ccw (priv->image);
-	g_object_unref (priv->image);
-	priv->image = rotated;
-
-	priv->modified = TRUE;
-	g_signal_emit (G_OBJECT (img), eog_image_signals [SIGNAL_IMAGE_CHANGED], 0);	
+		if (gtk_events_pending ()) {
+			gtk_main_iteration ();
+		}
+	}
 }
 
-void
-eog_image_rotate_180 (EogImage *img)
+static void
+image_transform (EogImage *img, EogTransform *trans, gboolean is_undo)
 {
 	EogImagePrivate *priv;
+	GdkPixbuf *transformed;
+	gboolean modified = FALSE;
 
 	g_return_if_fail (EOG_IS_IMAGE (img));
-	
-	priv = img->priv;
-	if (priv->image == NULL) return;
+	g_return_if_fail (EOG_IS_TRANSFORM (trans));
 
-	eog_pixbuf_rotate_180 (priv->image);
+	priv = img->priv;
+
+	if (priv->image != NULL) {
+		transformed = eog_transform_apply (trans, priv->image, transform_progress_hook_sync, img);
+		
+		g_object_unref (priv->image);
+		priv->image = transformed;
+       
+		modified = TRUE;
+	}
+
+	if (priv->thumbnail != NULL) {
+		transformed = eog_transform_apply (trans, priv->thumbnail, transform_progress_hook_sync, img);
+
+		g_object_unref (priv->thumbnail);
+		priv->thumbnail = transformed;
+       
+		modified = TRUE;
+	}
+
+	if (modified) {
+		priv->modified = TRUE;
+		g_signal_emit (G_OBJECT (img), eog_image_signals [SIGNAL_IMAGE_CHANGED], 0);
+	}
+
+	if (priv->trans == NULL) {
+		g_object_ref (trans);
+		priv->trans = trans;
+	}
+	else {
+		EogTransform *composition;
+
+		composition = eog_transform_compose (priv->trans, trans);
+
+		g_object_unref (priv->trans);
+		priv->trans = composition;
+	}
 	
-	priv->modified = TRUE;
-	g_signal_emit (G_OBJECT (img), eog_image_signals [SIGNAL_IMAGE_CHANGED], 0);
+	if (!is_undo) {
+		g_object_ref (trans);
+		priv->undo_stack = g_list_prepend (priv->undo_stack, trans);
+	}
 }
 
-void
-eog_image_flip_horizontal (EogImage *img)
+
+void                
+eog_image_transform (EogImage *img, EogTransform *trans)
 {
-	EogImagePrivate *priv;
-	
-	g_return_if_fail (EOG_IS_IMAGE (img));
+	gint signal_id;
 
-	priv = img->priv;
-	if (priv->image == NULL) return;
+	signal_id = g_timeout_add (CHECK_LOAD_TIMEOUT, check_progress_sync, img);
 
-	eog_pixbuf_flip_horizontal (priv->image);
+	image_transform (img, trans, FALSE);
 
-	priv->modified = TRUE;
-	g_signal_emit (G_OBJECT (img), eog_image_signals [SIGNAL_IMAGE_CHANGED], 0);	
+	g_source_remove (signal_id);
+
+	g_signal_emit (G_OBJECT (img), eog_image_signals [SIGNAL_PROGRESS], 0, 1.0);
 }
 
-void
-eog_image_flip_vertical (EogImage *img)
+void 
+eog_image_undo (EogImage *img)
 {
 	EogImagePrivate *priv;
-	
+	EogTransform *trans;
+	EogTransform *inverse;
+
 	g_return_if_fail (EOG_IS_IMAGE (img));
 
 	priv = img->priv;
-	if (priv->image == NULL) return;
 
-	eog_pixbuf_flip_vertical (priv->image);
+	if (priv->undo_stack != NULL) {
+		trans = EOG_TRANSFORM (priv->undo_stack->data);
 
-	priv->modified = TRUE;
-	g_signal_emit (G_OBJECT (img), eog_image_signals [SIGNAL_IMAGE_CHANGED], 0);	
+		inverse = eog_transform_reverse (trans);
+
+		image_transform (img, inverse, TRUE);
+
+		priv->undo_stack = g_list_delete_link (priv->undo_stack, priv->undo_stack);
+
+		if (eog_transform_is_identity (priv->trans)) {
+			g_object_unref (priv->trans);
+			priv->trans = NULL;
+		}
+	}
 }
 
 gboolean
@@ -1332,3 +1490,11 @@ eog_image_is_loaded (EogImage *img)
 	
 	return result;
 }
+
+GnomeVFSURI*
+eog_image_get_uri (EogImage *img)
+{
+	g_return_val_if_fail (EOG_IS_IMAGE (img), NULL);
+	
+	return gnome_vfs_uri_ref (img->priv->uri);
+}
