@@ -40,6 +40,10 @@
 #include "eog-full-screen.h"
 #include "eog-hig-dialog.h"
 
+#include <libgpi/gpi-dialog-pixbuf.h>
+#include <libgpi/gpi-mgr-pixbuf.h>
+#include <libgnomeprintui/gnome-print-job-preview.h>
+
 
 
 
@@ -85,6 +89,7 @@ enum {
 	PROP_CHECK_SIZE,
 	PROP_IMAGE_WIDTH,
 	PROP_IMAGE_HEIGHT,
+	PROP_IMAGE_PROGRESS,
 	PROP_WINDOW_WIDTH,
 	PROP_WINDOW_HEIGHT,
 	PROP_WINDOW_TITLE,
@@ -177,83 +182,57 @@ verb_FullScreen_cb (BonoboUIComponent *uic, gpointer data, const char *name)
 }
 
 static void
-verb_FlipHorizontal_cb (BonoboUIComponent *uic, gpointer data, const char *name)
+apply_transformation (EogImageView *view, EogTransform *trans)
 {
-	EogImageView *image_view;
 	EogImageViewPrivate *priv;
 
-	g_return_if_fail (EOG_IS_IMAGE_VIEW (data));
-
-	image_view = EOG_IMAGE_VIEW (data);
-	priv = image_view->priv;
+	priv = view->priv;
 
 	if (priv->image != NULL) {
-		eog_image_flip_horizontal (priv->image);
+		eog_image_transform (priv->image, trans);
 	}
+
+	g_object_unref (trans);
+}
+
+static void
+verb_FlipHorizontal_cb (BonoboUIComponent *uic, gpointer data, const char *name)
+{
+	g_return_if_fail (EOG_IS_IMAGE_VIEW (data));
+
+	apply_transformation (EOG_IMAGE_VIEW (data), eog_transform_flip_new (EOG_TRANSFORM_FLIP_HORIZONTAL));
 }
 
 static void
 verb_FlipVertical_cb (BonoboUIComponent *uic, gpointer data, const char *name)
 {
-	EogImageView *image_view;
-	EogImageViewPrivate *priv;
-
 	g_return_if_fail (EOG_IS_IMAGE_VIEW (data));
 
-	image_view = EOG_IMAGE_VIEW (data);
-	priv = image_view->priv;
-
-	if (priv->image != NULL) {
-		eog_image_flip_vertical (priv->image);
-	}
+	apply_transformation (EOG_IMAGE_VIEW (data), eog_transform_flip_new (EOG_TRANSFORM_FLIP_VERTICAL));
 }
 
 static void
 verb_Rotate90ccw_cb (BonoboUIComponent *uic, gpointer data, const char *name)
 {
-	EogImageView *image_view;
-	EogImageViewPrivate *priv;
-
 	g_return_if_fail (EOG_IS_IMAGE_VIEW (data));
 
-	image_view = EOG_IMAGE_VIEW (data);
-	priv = image_view->priv;
-
-	if (priv->image != NULL) {
-		eog_image_rotate_counter_clock_wise (priv->image);
-	}
+	apply_transformation (EOG_IMAGE_VIEW (data), eog_transform_rotate_new (270));
 }
 
 static void
 verb_Rotate90cw_cb (BonoboUIComponent *uic, gpointer data, const char *name)
 {
-	EogImageView *image_view;
-	EogImageViewPrivate *priv;
-
 	g_return_if_fail (EOG_IS_IMAGE_VIEW (data));
 
-	image_view = EOG_IMAGE_VIEW (data);
-	priv = image_view->priv;
-
-	if (priv->image != NULL) {
-		eog_image_rotate_clock_wise (priv->image);
-	}
+	apply_transformation (EOG_IMAGE_VIEW (data), eog_transform_rotate_new (90));
 }
 
 static void
 verb_Rotate180_cb (BonoboUIComponent *uic, gpointer data, const char *name)
 {
-	EogImageView *image_view;
-	EogImageViewPrivate *priv;
-
 	g_return_if_fail (EOG_IS_IMAGE_VIEW (data));
 
-	image_view = EOG_IMAGE_VIEW (data);
-	priv = image_view->priv;
-
-	if (priv->image != NULL) {
-		eog_image_rotate_180 (priv->image);
-	}
+	apply_transformation (EOG_IMAGE_VIEW (data), eog_transform_rotate_new (180));
 }
 
 static void
@@ -341,6 +320,62 @@ verb_SaveAs_cb (BonoboUIComponent *uic, gpointer user_data,
 	}
 }
 
+static void
+on_ok_clicked (GPIDialog *d, EogImageView *iv)
+{
+	gtk_idle_add ((GtkFunction) gtk_object_destroy, d);
+}
+
+static void
+on_cancel_clicked (GPIDialog *d, EogImageView *iv)
+{
+	gtk_idle_add ((GtkFunction) gtk_object_destroy, d);
+}
+
+static void
+verb_PrintSetup_cb (BonoboUIComponent *uic, gpointer user_data,
+		    const char *cname)
+{
+	EogImageView *iv = EOG_IMAGE_VIEW (user_data);
+	GdkPixbuf *pixbuf = eog_image_get_pixbuf (iv->priv->image);
+	GPIDialogPixbuf *d = gpi_dialog_pixbuf_new (pixbuf);
+
+	g_signal_connect (d, "ok_clicked", G_CALLBACK (on_ok_clicked), iv);
+	g_signal_connect (d, "cancel_clicked",
+			  G_CALLBACK (on_cancel_clicked), iv);
+	gtk_widget_show (GTK_WIDGET (d));
+}
+
+static void
+verb_PrintPreview_cb (BonoboUIComponent *uic, gpointer user_data,
+		      const char *cname)
+{
+	EogImageView *iv = EOG_IMAGE_VIEW (user_data);
+	GdkPixbuf *pixbuf = eog_image_get_pixbuf (iv->priv->image);
+	GPIMgrPixbuf *mgr = gpi_mgr_pixbuf_new (pixbuf);
+	GnomePrintJob *job = gpi_mgr_get_job (GPI_MGR (mgr));
+	GtkWidget *d;
+
+	d = gnome_print_job_preview_new (job, _("Preview"));
+	g_object_unref (job);
+	g_object_unref (mgr);
+	gtk_widget_show (d);
+}
+
+static void
+verb_Print_cb (BonoboUIComponent *uic, gpointer user_data,
+	       const char *cname)
+{
+	EogImageView *iv = EOG_IMAGE_VIEW (user_data);
+	GdkPixbuf *pixbuf = eog_image_get_pixbuf (iv->priv->image);
+	GPIMgrPixbuf *mgr = gpi_mgr_pixbuf_new (pixbuf);
+	GnomePrintJob *job = gpi_mgr_get_job (GPI_MGR (mgr));
+
+	gnome_print_job_print (job);
+	g_object_unref (job);
+	g_object_unref (mgr);
+}
+
 static BonoboUIVerb eog_zoom_verbs[] = {
 	BONOBO_UI_VERB ("ZoomIn",        verb_ZoomIn_cb),
 	BONOBO_UI_VERB ("ZoomOut",       verb_ZoomOut_cb),
@@ -357,6 +392,9 @@ static BonoboUIVerb eog_verbs[] = {
 	BONOBO_UI_VERB ("Rotate90cw",    verb_Rotate90cw_cb),
 	BONOBO_UI_VERB ("Rotate90ccw",   verb_Rotate90ccw_cb),
 	BONOBO_UI_VERB ("Rotate180",     verb_Rotate180_cb),
+	BONOBO_UI_VERB ("PrintSetup",    verb_PrintSetup_cb),
+	BONOBO_UI_VERB ("Print",         verb_Print_cb),
+	BONOBO_UI_VERB ("PrintPreview",  verb_PrintPreview_cb),
 	BONOBO_UI_VERB_END
 };
 
@@ -848,6 +886,27 @@ image_size_prepared_cb (EogImage *img, int width, int height, gpointer data)
 	bonobo_arg_release (arg);
 }
 
+static void
+image_progress_cb (EogImage *img, float progress, gpointer data)
+{
+	EogImageView *view;
+	EogImageViewPrivate *priv;
+	BonoboArg *arg;
+
+	g_return_if_fail (EOG_IS_IMAGE_VIEW (data));
+
+	view = EOG_IMAGE_VIEW (data);
+	priv = view->priv;
+
+	arg = bonobo_arg_new (BONOBO_ARG_FLOAT);
+	BONOBO_ARG_SET_FLOAT (arg, progress);
+
+	bonobo_event_source_notify_listeners (priv->property_bag->es,
+					      "image/progress",
+					      arg, NULL);
+	bonobo_arg_release (arg);
+}
+
 static gint
 load_uri_cb (BonoboPersistFile *pf, const CORBA_char *text_uri,
 	     CORBA_Environment *ev, void *closure)
@@ -863,6 +922,7 @@ load_uri_cb (BonoboPersistFile *pf, const CORBA_char *text_uri,
 
 	image = eog_image_new (text_uri);
 	g_signal_connect (image, "loading_size_prepared", G_CALLBACK (image_size_prepared_cb), view);
+	g_signal_connect (image, "progress", G_CALLBACK (image_progress_cb), view);
 
 	/* FIXME: remove potential signal handlers from old image object */
 	if (priv->image != NULL) {
@@ -1163,6 +1223,9 @@ eog_image_view_construct (EogImageView *image_view, gboolean need_close_item)
 	bonobo_property_bag_add (priv->property_bag, "window/height", PROP_WINDOW_HEIGHT,
 				 BONOBO_ARG_INT, NULL, _("Desired Window Height"),
 				 BONOBO_PROPERTY_READABLE);
+	bonobo_property_bag_add (priv->property_bag, "image/progress", PROP_IMAGE_PROGRESS,
+				 BONOBO_ARG_FLOAT, NULL, _("Progress of Image Loading"),
+				 BONOBO_PROPERTY_READABLE);
 	bonobo_control_set_properties (BONOBO_CONTROL (control),
 				       BONOBO_OBJREF (priv->property_bag),
 				       NULL);
